@@ -21,9 +21,7 @@ from collections import OrderedDict
 from flask import current_app, request, url_for
 from werkzeug.utils import import_string
 
-from flask_caching._compat import PY2, iteritems
-
-__version__ = "1.7.2"
+__version__ = "1.9.0"
 
 logger = logging.getLogger(__name__)
 
@@ -40,10 +38,16 @@ SUPPORTED_HASH_FUNCTIONS = [
 # Used to remove control characters and whitespace from cache keys.
 valid_chars = set(string.ascii_letters + string.digits + "_.")
 delchars = "".join(c for c in map(chr, range(256)) if c not in valid_chars)
-if PY2:
-    null_control = (None, delchars)
-else:
-    null_control = (dict((k, None) for k in delchars),)
+null_control = (dict((k, None) for k in delchars),)
+
+
+def wants_args(f):
+    """Check if the function wants any arguments
+    """
+
+    argspec = inspect.getfullargspec(f)
+
+    return bool(argspec.args or argspec.varargs or argspec.varkw)
 
 
 def get_arg_names(f):
@@ -52,40 +56,23 @@ def get_arg_names(f):
     :param f:
     :return: String list of arguments
     """
-    try:
-        # Python >= 3.3
-        sig = inspect.signature(f)
-        return [
-            parameter.name
-            for parameter in sig.parameters.values()
-            if parameter.kind == parameter.POSITIONAL_OR_KEYWORD
-        ]
-    except AttributeError:
-        try:
-            # Python >= 3.0
-            return inspect.getfullargspec(f).args
-        except AttributeError:
-            return inspect.getargspec(f).args
+    sig = inspect.signature(f)
+    return [
+        parameter.name
+        for parameter in sig.parameters.values()
+        if parameter.kind == parameter.POSITIONAL_OR_KEYWORD
+    ]
 
 
 def get_arg_default(f, position):
-    try:
-        # Python >= 3.3
-        sig = inspect.signature(f)
-        arg = list(sig.parameters.values())[position]
-        arg_def = arg.default
-        return arg_def if arg_def != inspect.Parameter.empty else None
-    except AttributeError:
-        try:
-            spec = inspect.getfullargspec(f)
-        except AttributeError:
-            spec = inspect.getargspec(f)
+    sig = inspect.signature(f)
+    arg = list(sig.parameters.values())[position]
+    arg_def = arg.default
+    return arg_def if arg_def != inspect.Parameter.empty else None
 
-        args_len = len(spec.args)
-        if spec.defaults and abs(position - args_len) <= len(spec.defaults):
-            return spec.defaults[position - args_len]
-        else:
-            return None
+
+def get_id(obj):
+    return getattr(obj, "__caching_id__", repr)(obj)
 
 
 def function_namespace(f, args=None):
@@ -97,9 +84,9 @@ def function_namespace(f, args=None):
     instance_self = getattr(f, "__self__", None)
 
     if instance_self and not inspect.isclass(instance_self):
-        instance_token = repr(f.__self__)
+        instance_token = get_id(f.__self__)
     elif m_args and m_args[0] == "self" and args:
-        instance_token = repr(args[0])
+        instance_token = get_id(args[0])
 
     module = f.__module__
 
@@ -162,6 +149,8 @@ class Cache(object):
         self.with_jinja2_ext = with_jinja2_ext
         self.config = config
 
+        self.source_check = None
+
         if app is not None:
             self.init_app(app, config)
 
@@ -193,6 +182,7 @@ class Cache(object):
         config.setdefault("CACHE_ARGS", [])
         config.setdefault("CACHE_TYPE", "null")
         config.setdefault("CACHE_NO_NULL_WARNING", False)
+        config.setdefault("CACHE_SOURCE_CHECK", False)
 
         if (
             config["CACHE_TYPE"] == "null"
@@ -203,6 +193,14 @@ class Cache(object):
                 "caching is effectively disabled."
             )
 
+        if config["CACHE_TYPE"] == "filesystem" and config["CACHE_DIR"] is None:
+            warnings.warn(
+                "Flask-Caching: CACHE_TYPE is set to filesystem but no "
+                "CACHE_DIR is set."
+            )
+
+        self.source_check = config["CACHE_SOURCE_CHECK"]
+
         if self.with_jinja2_ext:
             from .jinja2ext import CacheExtension, JINJA_CACHE_ATTR_NAME
 
@@ -281,6 +279,15 @@ class Cache(object):
         """Proxy function for internal cache object."""
         return self.cache.get_dict(*args, **kwargs)
 
+    def unlink(self, *args, **kwargs):
+        """Proxy function for internal cache object
+        only support Redis
+        """
+        unlink = getattr(self.cache, "unlink", None)
+        if unlink is not None and callable(unlink):
+            return unlink(*args, **kwargs)
+        return self.delete_many(*args, **kwargs)
+
     def cached(
         self,
         timeout=None,
@@ -290,6 +297,9 @@ class Cache(object):
         response_filter=None,
         query_string=False,
         hash_method=hashlib.md5,
+        cache_none=False,
+        make_cache_key=None,
+        source_check=None
     ):
         """Decorator. Use this to cache a function. By default the cache key
         is `view/request.path`. You are able to use this decorator with any
@@ -330,6 +340,8 @@ class Cache(object):
                 **make_cache_key**
                     A function used in generating the cache_key used.
 
+                    readable and writable
+
         :param timeout: Default None. If set to an integer, will cache for that
                         amount of time. Unit of time is in seconds.
 
@@ -372,7 +384,22 @@ class Cache(object):
 
         :param hash_method: Default hashlib.md5. The hash method used to
                             generate the keys for cached results.
+        :param cache_none: Default False. If set to True, add a key exists
+                           check when cache.get returns None. This will likely
+                           lead to wrongly returned None values in concurrent
+                           situations and is not recommended to use.
+        :param make_cache_key: Default None. If set to a callable object,
+                           it will be called to generate the cache key
 
+        :param source_check: Default None. If None will use the value set by
+                             CACHE_SOURCE_CHECK.
+                             If True, include the function's source code in the
+                             hash to avoid using cached values when the source
+                             code has changed and the input values remain the
+                             same. This ensures that the cache_key will be
+                             formed with the function's source code hash in
+                             addition to other parameters that may be included
+                             in the formation of the key.
         """
 
         def decorator(f):
@@ -382,15 +409,26 @@ class Cache(object):
                 if self._bypass_cache(unless, f, *args, **kwargs):
                     return f(*args, **kwargs)
 
-                try:
-                    if query_string:
-                        cache_key = _make_cache_key_query_string()
-                    else:
-                        cache_key = _make_cache_key(
-                            args, kwargs, use_request=True
-                        )
+                nonlocal source_check
+                if source_check is None:
+                    source_check = self.source_check
 
-                    if callable(forced_update) and forced_update() is True:
+                try:
+                    if make_cache_key is not None and callable(make_cache_key):
+                        cache_key = make_cache_key(*args, **kwargs)
+                    else:
+                        cache_key = _make_cache_key(args, kwargs, use_request=True)
+
+
+                    if (
+                        callable(forced_update)
+                        and (
+                            forced_update(*args, **kwargs)
+                            if wants_args(forced_update)
+                            else forced_update()
+                        )
+                        is True
+                    ):
                         rv = None
                         found = False
                     else:
@@ -401,13 +439,20 @@ class Cache(object):
                         # might be because the key is not found in the cache
                         # or because the cached value is actually None
                         if rv is None:
-                            found = self.cache.has(cache_key)
+                            # If we're sure we don't need to cache None values
+                            # (cache_none=False), don't bother checking for
+                            # key existence, as it can lead to false positives
+                            # if a concurrent call already cached the
+                            # key between steps. This would cause us to
+                            # return None when we shouldn't
+                            if not cache_none:
+                                found = False
+                            else:
+                                found = self.cache.has(cache_key)
                 except Exception:
                     if self.app.debug:
                         raise
-                    logger.exception(
-                        "Exception possibly due to cache backend."
-                    )
+                    logger.exception("Exception possibly due to cache backend.")
                     return f(*args, **kwargs)
 
                 if not found:
@@ -428,15 +473,11 @@ class Cache(object):
                             )
                 return rv
 
-            def make_cache_key(*args, **kwargs):
+            def default_make_cache_key(*args, **kwargs):
                 # Convert non-keyword arguments (which is the way
                 # `make_cache_key` expects them) to keyword arguments
                 # (the way `url_for` expects them)
-                try:
-                    # Python >= 3.0
-                    argspec_args = inspect.getfullargspec(f).args
-                except AttributeError:
-                    argspec_args = inspect.getargspec(f).args
+                argspec_args = inspect.getfullargspec(f).args
 
                 for arg_name, arg in zip(argspec_args, args):
                     kwargs[arg_name] = arg
@@ -449,6 +490,10 @@ class Cache(object):
                 Produces the same cache key regardless of argument order, e.g.,
                 both `?limit=10&offset=20` and `?offset=20&limit=10` will
                 always produce the same exact cache key.
+
+                If func is provided and is callable it will be used to hash
+                the function's source code and include it in the cache key.
+                This will only be done is source_check is True.
                 """
 
                 # Create a tuple of (key, value) pairs, where the key is the
@@ -457,6 +502,7 @@ class Cache(object):
                 # is always the same for query string args whose keys/values
                 # are the same, regardless of the order in which they are
                 # provided.
+
                 args_as_sorted_tuple = tuple(
                     sorted((pair for pair in request.args.items(multi=True)))
                 )
@@ -464,26 +510,48 @@ class Cache(object):
                 # used as a key for cache. Turn them into bytes so that the
                 # hash function will accept them
                 args_as_bytes = str(args_as_sorted_tuple).encode()
-                hashed_args = str(hash_method(args_as_bytes).hexdigest())
-                cache_key = request.path + hashed_args
+                cache_hash = hash_method(args_as_bytes)
+
+                # Use the source code if source_check is True and update the
+                # cache_hash before generating the hashing and using it in
+                # cache_key
+                if source_check and callable(f):
+                    func_source_code = inspect.getsource(f)
+                    cache_hash.update(func_source_code.encode("utf-8"))
+
+                cache_hash = str(cache_hash.hexdigest())
+
+                cache_key = request.path + cache_hash
+
                 return cache_key
 
             def _make_cache_key(args, kwargs, use_request):
-                if callable(key_prefix):
-                    cache_key = key_prefix()
-                elif "%s" in key_prefix:
-                    if use_request:
-                        cache_key = key_prefix % request.path
-                    else:
-                        cache_key = key_prefix % url_for(f.__name__, **kwargs)
+                if query_string:
+                    return _make_cache_key_query_string()
                 else:
-                    cache_key = key_prefix
+                    if callable(key_prefix):
+                        cache_key = key_prefix()
+                    elif "%s" in key_prefix:
+                        if use_request:
+                            cache_key = key_prefix % request.path
+                        else:
+                            cache_key = key_prefix % url_for(f.__name__, **kwargs)
+                    else:
+                        cache_key = key_prefix
+
+                if source_check and callable(f):
+                    func_source_code = inspect.getsource(f)
+                    func_source_hash = hash_method(
+                        func_source_code.encode("utf-8"))
+                    func_source_hash = str(func_source_hash.hexdigest())
+
+                    cache_key += func_source_hash
 
                 return cache_key
 
             decorated_function.uncached = f
             decorated_function.cache_timeout = timeout
-            decorated_function.make_cache_key = make_cache_key
+            decorated_function.make_cache_key = default_make_cache_key
 
             return decorated_function
 
@@ -499,6 +567,7 @@ class Cache(object):
         self,
         f,
         args=None,
+        kwargs=None,
         reset=False,
         delete=False,
         timeout=None,
@@ -524,7 +593,15 @@ class Cache(object):
         version_data_list = list(self.cache.get_many(*fetch_keys))
         dirty = False
 
-        if callable(forced_update) and forced_update() is True:
+        if (
+            callable(forced_update)
+            and (
+                forced_update(*(args or ()), **(kwargs or {}))
+                if wants_args(forced_update)
+                else forced_update()
+            )
+            is True
+        ):
             # Mark key as dirty to update its TTL
             dirty = True
 
@@ -556,6 +633,7 @@ class Cache(object):
         timeout=None,
         forced_update=False,
         hash_method=hashlib.md5,
+        source_check=False
     ):
         """Function used to create the cache_key for memoized functions."""
 
@@ -580,6 +658,13 @@ class Cache(object):
 
             cache_key = hash_method()
             cache_key.update(updated.encode("utf-8"))
+
+            # Use the source code if source_check is True and update the
+            # cache_key with the function's source.
+            if source_check and callable(f):
+                func_source_code = inspect.getsource(f)
+                cache_key.update(func_source_code.encode("utf-8"))
+
             cache_key = base64.b64encode(cache_key.digest())[:16]
             cache_key = cache_key.decode("utf-8")
             cache_key += version_data
@@ -605,11 +690,11 @@ class Cache(object):
         for i in range(args_len):
             arg_default = get_arg_default(f, i)
             if i == 0 and arg_names[i] in ("self", "cls"):
-                #: use the repr of the class instance
+                #: use the id func of the class instance
                 #: this supports instance methods for
                 #: the memoized functions, giving more
                 #: flexibility to developers
-                arg = repr(args[0])
+                arg = get_id(args[0])
                 arg_num += 1
             elif arg_names[i] in kwargs:
                 arg = kwargs[arg_names[i]]
@@ -632,7 +717,7 @@ class Cache(object):
             #     try:
             #         arg = hash(arg)
             #     except:
-            #         arg = repr(arg)
+            #         arg = get_id(arg)
 
             #: Or what about a special __cacherepr__ function
             #: on an object, this allows objects to act normal
@@ -645,14 +730,12 @@ class Cache(object):
 
             new_args.append(arg)
 
-        new_args.extend(args[len(arg_names):])
+        new_args.extend(args[len(arg_names) :])
         return (
             tuple(new_args),
             OrderedDict(
                 sorted(
-                    (k, v)
-                    for k, v in iteritems(kwargs)
-                    if k in kw_keys_remaining
+                    (k, v) for k, v in kwargs.items() if k in kw_keys_remaining
                 )
             ),
         )
@@ -665,21 +748,8 @@ class Cache(object):
         bypass_cache = False
 
         if callable(unless):
-            try:
-                # Python >= 3.0
-                argspec = inspect.getfullargspec(unless)
-                has_args = (
-                    len(argspec.args) > 0 or
-                    argspec.varargs or
-                    argspec.varkw
-                )
-            except AttributeError:
-                argspec = inspect.getargspec(unless)
-                has_args = (
-                    len(argspec.args) > 0 or
-                    argspec.varargs or
-                    argspec.keywords
-                )
+            argspec = inspect.getfullargspec(unless)
+            has_args = len(argspec.args) > 0 or argspec.varargs or argspec.varkw
 
             # If unless() takes args, pass them in.
             if has_args:
@@ -696,7 +766,10 @@ class Cache(object):
         make_name=None,
         unless=None,
         forced_update=None,
+        response_filter=None,
         hash_method=hashlib.md5,
+        cache_none=False,
+        source_check=None
     ):
         """Use this to cache the result of a function, taking its arguments
         into account in the cache key.
@@ -752,8 +825,28 @@ class Cache(object):
                               cache value will be updated regardless cache
                               is expired or not. Useful for background
                               renewal of cached functions.
+        :param response_filter: Default None. If not None, the callable is
+                                invoked after the cached funtion evaluation,
+                                and is given one arguement, the response
+                                content. If the callable returns False, the
+                                content will not be cached. Useful to prevent
+                                caching of code 500 responses.
         :param hash_method: Default hashlib.md5. The hash method used to
                             generate the keys for cached results.
+        :param cache_none: Default False. If set to True, add a key exists
+                           check when cache.get returns None. This will likely
+                           lead to wrongly returned None values in concurrent
+                           situations and is not recommended to use.
+
+        :param source_check: Default None. If None will use the value set by
+                             CACHE_SOURCE_CHECK.
+                             If True, include the function's source code in the
+                             hash to avoid using cached values when the source
+                             code has changed and the input values remain the
+                             same. This ensures that the cache_key will be
+                             formed with the function's source code hash in
+                             addition to other parameters that may be included
+                             in the formation of the key.
 
         .. versionadded:: 0.5
             params ``make_name``, ``unless``
@@ -766,12 +859,24 @@ class Cache(object):
                 if self._bypass_cache(unless, f, *args, **kwargs):
                     return f(*args, **kwargs)
 
+                nonlocal source_check
+                if source_check is None:
+                    source_check = self.source_check
+
                 try:
                     cache_key = decorated_function.make_cache_key(
                         f, *args, **kwargs
                     )
 
-                    if callable(forced_update) and forced_update() is True:
+                    if (
+                        callable(forced_update)
+                        and (
+                            forced_update(*args, **kwargs)
+                            if wants_args(forced_update)
+                            else forced_update()
+                        )
+                        is True
+                    ):
                         rv = None
                         found = False
                     else:
@@ -782,29 +887,38 @@ class Cache(object):
                         # might be because the key is not found in the cache
                         # or because the cached value is actually None
                         if rv is None:
-                            found = self.cache.has(cache_key)
+                            # If we're sure we don't need to cache None values
+                            # (cache_none=False), don't bother checking for
+                            # key existence, as it can lead to false positives
+                            # if a concurrent call already cached the
+                            # key between steps. This would cause us to
+                            # return None when we shouldn't
+                            if not cache_none:
+                                found = False
+                            else:
+                                found = self.cache.has(cache_key)
                 except Exception:
                     if self.app.debug:
                         raise
-                    logger.exception(
-                        "Exception possibly due to cache backend."
-                    )
+                    logger.exception("Exception possibly due to cache backend.")
                     return f(*args, **kwargs)
 
                 if not found:
                     rv = f(*args, **kwargs)
-                    try:
-                        self.cache.set(
-                            cache_key,
-                            rv,
-                            timeout=decorated_function.cache_timeout,
-                        )
-                    except Exception:
-                        if self.app.debug:
-                            raise
-                        logger.exception(
-                            "Exception possibly due to cache backend."
-                        )
+
+                    if response_filter is None or response_filter(rv):
+                        try:
+                            self.cache.set(
+                                cache_key,
+                                rv,
+                                timeout=decorated_function.cache_timeout,
+                            )
+                        except Exception:
+                            if self.app.debug:
+                                raise
+                            logger.exception(
+                                "Exception possibly due to cache backend."
+                            )
                 return rv
 
             decorated_function.uncached = f
@@ -814,6 +928,7 @@ class Cache(object):
                 timeout=decorated_function,
                 forced_update=forced_update,
                 hash_method=hash_method,
+                source_check=source_check
             )
             decorated_function.delete_memoized = lambda: self.delete_memoized(f)
 
@@ -894,9 +1009,9 @@ class Cache(object):
             3.72341788
 
         :param fname: The memoized function.
-        :param \*args: A list of positional parameters used with
+        :param \\*args: A list of positional parameters used with
                        memoized function.
-        :param \**kwargs: A dict of named parameters used with
+        :param \\**kwargs: A dict of named parameters used with
                           memoized function.
 
         .. note::
